#define MAX_DISPARITY 64
#define LARGE_VALUE 100
#define HUGE_VALUE 16777216

typedef Graph<float,float,float> GraphType;


//Calculate data cost
double cal_data_Cost(int,int);
//Calculate smooth cost
double cal_smooth_cost(uchar,uchar,uchar,
			      uchar,uchar,uchar);
//disparity
int **disparity_Array(int);

//Calculate alpha node
int cal_alpha_Node(int,int,int,int,int);
//Calculate SAD
double cal_SAD(uchar**,uchar**,int,int,int,int);

//blockmaching
void block_matching(uchar**,uchar**,uchar**,int,int,int,int);
//alpha-expansion
void alpha_expansion(uchar**,uchar**,uchar**,int,int,int,int);
//hierarchical
void hierarchical(uchar**,uchar**,uchar**,int,int,int,int,int**);


double cal_data_Cost(double a)
{
  double weight;

  weight=0.5;

  return weight*a;
}

double cal_smooth_cost(double a,double b)
{
  double weight;

  if(fabs(a-b)<0.000000001)
    return 0.0;

  weight=3.0;

  return weight*fabs(a-b);
}

int cal_alpha_Node(int flag,int x,int y,
			  int img_width,int img_height)
{
  int value=0;

  switch(flag)
    {
    case 0:
      {
	value=img_width*y+x;
	break;
      }
    case 1://x 
      {
	value=(img_width*img_height)+
	  ((img_width-1)*y)+x;
	break;
      }
    case 2://y
      {
	value=(img_width*img_height)+
	  ((img_width-1)*img_height)+
	  ((img_height-1)*x)+y;
	break;
      }
    default:
      break;
    }

  return value;
}

int **disparity_Array(int n)
{
  int **value,count;

  value=alloc_2Di(n,n);

  for(int j=0;j<n;j++)
    for(int i=0;i<n;i++)  
      value[j][i]=-1;

  value[0][0]=0;
  value[1][0]=n/2;

  for(int j=2;j<=11;j++)
    {
      count=0;
      for(int i=0;i<(int)pow(2,j/2-1);i++)
	{
	  if(j%2==0)
	    value[j][i]=(int)((1+4*i)/pow(2,j/2+1)*n);
	  else
	    value[j][i]=(int)((3+4*i)/pow(2,j/2+1)*n);
	}
    }

  return value;
}

double cal_SAD(uchar **left,uchar **right,
		      int img_width,int img_height,int x,int y,int d,int blocksize)
{

  double error=0.0;
  int count=0;

  for(int j=-blocksize;j<=blocksize;j++)
    for(int i=-blocksize;i<=blocksize;i++)  
      {
	if(y+j<0             || 
	   y+j>img_height-1  || 
	   x+i-d<0           || 
	   x+i-d>img_width-1 ||
	   (x==0 && y==0)
	   )
	  continue;

	error+=
	  sqrt( pow( ( (double)left[y+j][x+i] - (double)right[y+j][x+i-d] ),2.0) );
	 
	count++;
      }
  
  if(count>0)
    return error/=(double)count;
  else
    return LARGE_VALUE;
}

void block_matching(uchar **left,uchar **right,uchar **out,
			   int img_width,int img_height,int max_disparity,int blocksize)
{

  int tmp_disparity=0;
  int tmp_count;
  double min,error;

  blocksize/=2;

  for(int j=0;j<img_height;j++)
    for(int i=0;i<img_width;i++)  
      {
	min = 16777216.0;

	for(int l=0;l<max_disparity;l++)  
	  {
	    error=cal_SAD(left,right,img_width,img_height,
				     i,j,l,blocksize);

	    if(error<min)
	      {
		min = error;
		tmp_disparity=l;
	      }
	  }
	out[j][i]=(uchar)(tmp_disparity*(255.0/(double)max_disparity));
      }
}

void alpha_expansion(uchar **left,uchar **right,uchar **out,
			    int img_width,int img_height,int max_disparity,int blocksize)
{
  int tmp_disparity=0;
  int tmp_count;
 
  GraphType *graph;

  uchar **default_cost,**tmp_cost;
  double **data_a,**data_b;
  double smooth_aa,smooth_bb,smooth_ab,smooth_ba;

  double min_flow,calc_flow;
  uchar flag;

  blocksize/=2;

  min_flow=HUGE_VALUE;

  //first data cost
  default_cost = alloc_2Dub(img_width,img_height);
  tmp_cost     = alloc_2Dub(img_width,img_height);
  data_a       = alloc_2Dd(img_width,img_height);
  data_b       = alloc_2Dd(img_width,img_height);


  block_matching(left,right,tmp_cost,
			img_width,img_height,max_disparity,blocksize*2);  

  for(int j=0;j<img_height;j++)
    for(int i=0;i<img_width;i++)    
      default_cost[j][i]=tmp_cost[j][i]/(192/max_disparity);


  //Ȥι̵ʤޤǷ֤
  for(int h=0;h<HUGE_VALUE;h++)
    {
      //flag
      flag=1;

      //ƻ뺹ˤĤƷ֤
      for(int g=0;g<max_disparity;g++)
	{
	  //դ
	  graph=new GraphType(img_width*img_height + (img_width-1)*img_height,
			      img_width*(img_height-1) + ((img_width-1)*img_height)*2 
			      + (img_width*(img_height-1))*2);

	  //Ρɤɲ
	  graph->add_node( img_width*img_height 
			   + (img_width-1)*img_height
			   +  img_width*(img_height-1));

	  //cal data cost
	  for(int i=0;i<img_height;i++)
	    for(int j=0;j<img_width;j++)  
	      {
		data_a[i][j]=cal_data_Cost(cal_SAD(left,right,img_width,img_height,
								 j,i,g,blocksize));
		data_b[i][j]=cal_data_Cost(cal_SAD(left,right,img_width,img_height,
								 j,i,default_cost[i][j],blocksize));

		graph->add_tweights(cal_alpha_Node(0,j,i,img_width,img_height),
				    data_b[i][j],data_a[i][j]);
	      }

	  //x smooth cost
	  for(int i=0;i<img_height;i++)
	    for(int j=0;j<img_width-1;j++)
	      {
		smooth_aa=cal_smooth_cost(g,g);
		smooth_bb=cal_smooth_cost(default_cost[i][j],default_cost[i][j+1]);
		smooth_ab=cal_smooth_cost(g,default_cost[i][j+1]);
		smooth_ba=cal_smooth_cost(default_cost[i][j],g);

		//x data cost
		graph->add_tweights(cal_alpha_Node(1,j,i,img_width,img_height),
				    smooth_bb,smooth_aa);

		if(smooth_bb<=smooth_aa)
		  {
		    graph->add_edge(cal_alpha_Node(0,j,i,img_width,img_height),
				    cal_alpha_Node(1,j,i,img_width,img_height),
				    Max(0,smooth_ab-smooth_bb),
				    LARGE_VALUE);
		    graph->add_edge(cal_alpha_Node(1,j,i,img_width,img_height),
				    cal_alpha_Node(0,j+1,i,img_width,img_height),
				    LARGE_VALUE,
				    Max(0,smooth_ba-smooth_bb));
		  }
		else if(smooth_bb>smooth_aa)
		  {
		    graph->add_edge(cal_alpha_Node(0,j,i,img_width,img_height),
				    cal_alpha_Node(1,j,i,img_width,img_height),
				    LARGE_VALUE,
				    Max(0,smooth_ba-smooth_aa));
		    graph->add_edge(cal_alpha_Node(1,j,i,img_width,img_height),
				    cal_alpha_Node(0,j+1,i,img_width,img_height),
				    Max(0,smooth_ab-smooth_aa),
				    LARGE_VALUE);
		  }
	      }
	  
	  //y smooth cost
	  for(int i=0;i<img_height-1;++i)
	    for(int j=0;j<img_width;++j)
	      {
	
		smooth_aa=cal_smooth_cost(g,g);
		smooth_bb=cal_smooth_cost(default_cost[i][j],default_cost[i+1][j]);
		smooth_ab=cal_smooth_cost(g,default_cost[i+1][j]);
		smooth_ba=cal_smooth_cost(default_cost[i][j],g);

		//y data cost
		graph->add_tweights(cal_alpha_Node(2,j,i,img_width,img_height),
				    smooth_bb,smooth_aa);

		if(smooth_bb<=smooth_aa)
		  {
		    graph->add_edge(cal_alpha_Node(0,j,i,img_width,img_height),
				    cal_alpha_Node(2,j,i,img_width,img_height),
				    Max(0,smooth_ab-smooth_bb),
				    LARGE_VALUE);
		    graph->add_edge(cal_alpha_Node(2,j,i,img_width,img_height),
				    cal_alpha_Node(0,j,i+1,img_width,img_height),
				    LARGE_VALUE,
				    Max(0,smooth_ba-smooth_bb));
		  }
		else if(smooth_bb>smooth_aa)
		  {
		    graph->add_edge(cal_alpha_Node(0,j,i,img_width,img_height),
				    cal_alpha_Node(2,j,i,img_width,img_height),
				    LARGE_VALUE,
				    Max(0,smooth_ba-smooth_aa));
		    graph->add_edge(cal_alpha_Node(2,j,i,img_width,img_height),
				    cal_alpha_Node(0,j,i+1,img_width,img_height),
				    Max(0,smooth_ab-smooth_aa),
				    LARGE_VALUE);
		  }
	      }

	  //****ήǾåȤη׻****//
	  calc_flow=graph->maxflow();

	  if(calc_flow<min_flow)
	    {
	      min_flow=calc_flow;
	      flag=0;

	      //****٥򹹿****//
	      for(int j=0;j<img_height;j++)
		for(int i=0;i<img_width;i++)
		  if(graph->what_segment(cal_alpha_Node(0,i,j,img_width,img_height)) == GraphType::SOURCE)
		    default_cost[j][i]=g;
	    }
	  
	  //****դξõ****//
	  delete(graph);
	  
	}

      if(flag==1)
	break;
    }

  for(int j=0;j<img_height;j++)
    for(int i=0;i<img_width;i++)
      out[j][i]=(uchar)(default_cost[j][i]*(192.0/(double)max_disparity));
}
	

void hierarchical(uchar **left,uchar **right,uchar **out,
			 int img_width,int img_height,int max_disparity,int blocksize,
			 int **disparity_array)
{
  int tmp_disparity=0;
  int tmp_count;

  GraphType *graph;

  uchar **default_cost,**tmp_cost;
  double **data_a,**data_b;
  double smooth_aa,smooth_bb,smooth_ab,smooth_ba;

  int **tmp_g;

  double min_flow,calc_flow;
  uchar flag;

  blocksize/=2;

  min_flow=HUGE_VALUE;

  //first data cost
  default_cost = alloc_2Dub(img_width,img_height);
  tmp_cost     = alloc_2Dub(img_width,img_height);
  data_a       = alloc_2Dd(img_width,img_height);
  data_b       = alloc_2Dd(img_width,img_height);
  tmp_g        = alloc_2Di(img_width,img_height);


  block_matching(left,right,tmp_cost,
			img_width,img_height,max_disparity,blocksize*2);  

  for(int j=0;j<img_height;j++)
    for(int i=0;i<img_width;i++)    
      default_cost[j][i]=tmp_cost[j][i]/(192/max_disparity);


  //Ȥι̵ʤޤǷ֤
  for(int h=0;h<HUGE_VALUE;h++)
    {
      //flag
      flag=1;

      //ƻ뺹ˤĤƷ֤
      for(int g=0;g<12;g++)
	{
	  //դ
	  graph=new GraphType(img_width*img_height + (img_width-1)*img_height,
			      img_width*(img_height-1) + ((img_width-1)*img_height)*2+
			      (img_width*(img_height-1))*2);

	  //Ρɤɲ
	  graph->add_node(img_width*img_height
			  + (img_width-1)*img_height
			  + img_width*(img_height-1));

	  //cal data cost
	  for(int i=0;i<img_height;i++)
	    for(int j=0;j<img_width;j++)  
	      {
		tmp_g[i][j]=HUGE_VALUE;
		for(int f=0;disparity_array[g][f]!=-1;++f)
		  if(abs(disparity_array[g][f]-default_cost[i][j])<
		     abs(tmp_g[i][j]-default_cost[i][j]))
		    tmp_g[i][j]=disparity_array[g][f];

		data_a[i][j]=cal_data_Cost(cal_SAD(left,right,img_width,img_height,
								 j,i,tmp_g[i][j],blocksize));
		data_b[i][j]=cal_data_Cost(cal_SAD(left,right,img_width,img_height,
								 j,i,default_cost[i][j],blocksize));

		graph->add_tweights(cal_alpha_Node(0,j,i,img_width,img_height),
				    data_b[i][j],data_a[i][j]);
	      }

	  //x smooth cost
	  for(int i=0;i<img_height;++i)
	    for(int j=0;j<img_width-1;++j)
	      {
		

		smooth_aa=cal_smooth_cost(tmp_g[i][j],tmp_g[i][j+1]);
		smooth_bb=cal_smooth_cost(default_cost[i][j],default_cost[i][j+1]);
		smooth_ab=cal_smooth_cost(tmp_g[i][j],default_cost[i][j+1]);
		smooth_ba=cal_smooth_cost(default_cost[i][j],tmp_g[i][j+1]);

		//x data cost
		graph->add_tweights(cal_alpha_Node(1,j,i,img_width,img_height),
				    smooth_bb,smooth_aa);

		if(smooth_bb<=smooth_aa)
		  {
		    graph->add_edge(cal_alpha_Node(0,j,i,img_width,img_height),
				    cal_alpha_Node(1,j,i,img_width,img_height),
				    Max(0,smooth_ab-smooth_bb),
				    LARGE_VALUE);
		    graph->add_edge(cal_alpha_Node(1,j,i,img_width,img_height),
				    cal_alpha_Node(0,j+1,i,img_width,img_height),
				    LARGE_VALUE,
				    Max(0,smooth_ba-smooth_bb));
		  }
		else if(smooth_bb>smooth_aa)
		  {
		    graph->add_edge(cal_alpha_Node(0,j,i,img_width,img_height),
				    cal_alpha_Node(1,j,i,img_width,img_height),
				    LARGE_VALUE,
				    Max(0,smooth_ba-smooth_aa));
		    graph->add_edge(cal_alpha_Node(1,j,i,img_width,img_height),
				    cal_alpha_Node(0,j+1,i,img_width,img_height),
				    Max(0,smooth_ab-smooth_aa),
				    LARGE_VALUE);
		  }
	      }
	  
	  //y smooth cost
	  for(int i=0;i<img_height-1;i++)
	    for(int j=0;j<img_width;j++)
	      {
		

		smooth_aa=cal_smooth_cost(tmp_g[i][j],tmp_g[i+1][j]);
		smooth_bb=cal_smooth_cost(default_cost[i][j],default_cost[i+1][j]);
		smooth_ab=cal_smooth_cost(tmp_g[i][j],default_cost[i+1][j]);
		smooth_ba=cal_smooth_cost(default_cost[i][j],tmp_g[i+1][j]);

		//y data cost
		graph->add_tweights(cal_alpha_Node(2,j,i,img_width,img_height),
				    smooth_bb,smooth_aa);

		if(smooth_bb<=smooth_aa)
		  {
		    graph->add_edge(cal_alpha_Node(0,j,i,img_width,img_height),
				    cal_alpha_Node(2,j,i,img_width,img_height),
				    Max(0,smooth_ab-smooth_bb),
				    LARGE_VALUE);
		    graph->add_edge(cal_alpha_Node(2,j,i,img_width,img_height),
				    cal_alpha_Node(0,j,i+1,img_width,img_height),
				    LARGE_VALUE,
				    Max(0,smooth_ba-smooth_bb));
		  }
		else if(smooth_bb>smooth_aa)
		  {
		    graph->add_edge(cal_alpha_Node(0,j,i,img_width,img_height),
				    cal_alpha_Node(2,j,i,img_width,img_height),
				    LARGE_VALUE,
				    Max(0,smooth_ba-smooth_aa));
		    graph->add_edge(cal_alpha_Node(2,j,i,img_width,img_height),
				    cal_alpha_Node(0,j,i+1,img_width,img_height),
				    Max(0,smooth_ab-smooth_aa),
				    LARGE_VALUE);
		  }
	      }

	  //****ήǾåȤη׻****//
	  calc_flow=graph->maxflow();

	  if(calc_flow<min_flow)
	    {
	      min_flow=calc_flow;
	      flag=0;

	      //****٥򹹿****//
	      for(int j=0;j<img_height;j++)
		for(int i=0;i<img_width;i++)
		  if(graph->what_segment(cal_alpha_Node(0,i,j,img_width,img_height)) == GraphType::SOURCE)
		    default_cost[j][i]=tmp_g[j][i];
	    }
	  
	  //****դξõ****//
	  delete(graph);

	}

      if(flag==1)
	break;
    }

  for(int j=0;j<img_height;j++)
    for(int i=0;i<img_width;i++)
      out[j][i]=(uchar)(default_cost[j][i]*(192.0/(double)max_disparity));  
}
